#include "StdAfx.h"
#include ".\component.h"

#include "Container.h"
#include "RootContainer.h"
#include "MouseEvent.h"
#include "Binding.h"

IMPLEMENT_ROOT_RTTI(CComponent)

//properties
LPCTSTR ComponentProperty::MouseEnter = _T("mouseenter");
LPCTSTR ComponentProperty::MouseExit = _T("mouseexit");
LPCTSTR ComponentProperty::OnCreate = _T("oncreate");
LPCTSTR ComponentProperty::OnDestroy = _T("ondestroy");

CComponent::CComponent(void)
{
	m_rcBound.left = m_rcBound.right = m_rcBound.top = m_rcBound.bottom = LONG_MAX;
	m_rcActualBound.left = m_rcActualBound.right = m_rcActualBound.top = m_rcActualBound.bottom = 0;
	ZeroMemory(m_szName, sizeof(TCHAR) * MAX_COMPONENT_NAME);
	m_pParent = NULL;
	m_bEnter = FALSE;
	m_pComponentUI = NULL;
	m_bVisible = TRUE;
	m_bEnabled = TRUE;
	m_hGetDCRgn = NULL;
	m_bEventRouteFlag = TRUE;
	m_HorizontalAlignment = HorizonAlignment::Stretch;
	m_VerticalAlignment = VerticalAlignment::Stretch;
	m_Margin.left = m_Margin.top = m_Margin.right = m_Margin.bottom = 0;

	static int s_lComponent = 0;
	USES_CONVERSION;
	_stprintf_s(m_szName, _T("%s-%d"), A2CT(GetRTTI()->GetClassName()), s_lComponent++);	
}

CComponent::~CComponent(void)
{
	_DestroyBinding();
	OnDestroy();
	m_pComponentUI = NULL;
}
void CComponent::Create()
{	
	this->OnCreate();
	CPropertyChangeEvent PropEvent(this, ComponentProperty::OnCreate);
	this->FirePropertyChangedEvent(&PropEvent);
}

void CComponent::Destroy()
{	
	this->OnDestroy();
	CPropertyChangeEvent PropEvent(this, ComponentProperty::OnDestroy);
	this->FirePropertyChangedEvent(&PropEvent);

	_RemoveParent();
}

BOOL CComponent::MoveWindow(int nX, int nY, int nWidth, int nHeight, BOOL bRepaint)
{
	RECT rcNew;
	rcNew.left = nX;
	rcNew.top = nY;
	rcNew.right = (nWidth==LONG_MAX) ? nWidth : GetRealValue(nX) + nWidth;
	rcNew.bottom = (nHeight==LONG_MAX) ? nHeight : GetRealValue(nY) + nHeight;
	CRect rcBoundNew;
	
	return MoveWindow(&rcNew, bRepaint);
}
BOOL CComponent::MoveWindow(LPRECT lpRect, BOOL bRepaint)
{
	CRect rcBoundOld;
	GetBounds(&rcBoundOld);

	CopyRect(&m_rcBound, lpRect);
	
	CSize sizePreferred(0, 0);
	if (this->GetPreferredSize(&sizePreferred) == FALSE)
	{
		sizePreferred.SetSize(0, 0);
	}

	m_rcActualBound.left = GetRealValue(m_rcBound.left);
	m_rcActualBound.top = GetRealValue(m_rcBound.top);
	m_rcActualBound.right = GetRealValue(m_rcBound.right, m_rcActualBound.left + sizePreferred.cx);
	m_rcActualBound.bottom = GetRealValue(m_rcBound.bottom, m_rcActualBound.top + sizePreferred.cy);
	
	CRect rcBoundNew;
	GetBounds(&rcBoundNew);

	this->OnSize(rcBoundNew.Width(), rcBoundNew.Height());

	this->_Monitor_CursorEnterExit(&rcBoundOld, &rcBoundNew);
	return TRUE;
}
BOOL CComponent::SetActualBound(int nX, int nY, int nWidth, int nHeight)
{
	m_rcActualBound.left = nX;
	m_rcActualBound.top = nY;
	m_rcActualBound.right = nX + nWidth;
	m_rcActualBound.bottom = nY + nHeight;

	CRect rcBoundNew;
	GetBounds(&rcBoundNew);
	this->OnSize(rcBoundNew.Width(), rcBoundNew.Height());
	
	return TRUE;
}
BOOL CComponent::GetClientRect(LPRECT lpRect)
{
	BOOL bSuccess = FALSE;
	if (lpRect)
	{
		CRect rcBound;
		GetBounds(&rcBound);

		bSuccess = CopyRect(lpRect, rcBound);
		OffsetRect(lpRect, -rcBound.left, -rcBound.top);
	}
	return bSuccess;
}
BOOL CComponent::GetWindowRect(LPRECT lpRect)
{
	if (NULL == lpRect) return FALSE;

	CRootContainer* pRoot = GetRootContainer();
	if (pRoot == NULL) return FALSE;
	
	CRect rcBound;
	GetBounds(&rcBound);
	CRect rcWinRect(rcBound);
	POINT ptAbsolute;
	GetParentAbsolutePosition(&ptAbsolute);
	rcWinRect.OffsetRect(ptAbsolute);
	
	CWindow wndOwner = pRoot->GetOwner();
	ATLASSERT(wndOwner.IsWindow());
	wndOwner.ClientToScreen(&rcWinRect);
	//CRect rcWin;
	//::GetWindowRect(hWndOwner, rcWin);
	//rcWinRect.OffsetRect(rcWin.TopLeft());
	
	BOOL bSuccess = CopyRect(lpRect, rcWinRect);
	
	return bSuccess;
}
BOOL CComponent::GetBounds(LPRECT lpRect)
{
	BOOL bSuccess = FALSE;
	if (lpRect)
	{				
		bSuccess = CopyRect(lpRect, &m_rcActualBound);			
	}
	return bSuccess;
}
BOOL CComponent::GetOriginalBounds(LPRECT lpRect)
{
	BOOL bSuccess = FALSE;
	if (lpRect)
	{				
		bSuccess = CopyRect(lpRect, &m_rcBound);
	}
	return bSuccess;
}
BOOL CComponent::GetDrawingBounds(LPRECT lpRect)
{
	BOOL bSuccess = FALSE;
	if (lpRect)
	{				
		//bSuccess = GetBounds(lpRect);
		bSuccess = GetClientRect(lpRect);
		//lpRect->right -= (m_Margin.right);
		//lpRect->bottom -= (m_Margin.bottom);
	}
	return bSuccess;
}
BOOL CComponent::InvalidateRect(LPCRECT pRect, BOOL bErase)
{
	if (IsVisible() == FALSE) return TRUE;

	CRootContainer* pRoot = GetRootContainer();
	if (pRoot == NULL) return FALSE;

	HWND hWndOwner = pRoot->GetOwner();
	if (IsWindow(hWndOwner) == FALSE) return FALSE;
	ATLASSERT(IsWindow(hWndOwner));

	CRect rcBound;
	GetBounds(&rcBound);

	CRect rcRepaint(pRect!=NULL ? pRect : rcBound);
	POINT ptAbsolute;
	GetParentAbsolutePosition(&ptAbsolute);
	rcRepaint.OffsetRect(ptAbsolute);	

	return ::InvalidateRect(hWndOwner, rcRepaint, bErase);
}
BOOL CComponent::EnableWindow(BOOL bEnabled)
{
	m_bEnabled = bEnabled;
	return TRUE;
}
BOOL CComponent::IsWindowEnabled()
{
	return m_bEnabled;
}
BOOL CComponent::ClientToScreen(LPPOINT lpPoint)
{
	CRootContainer* pRoot = GetRootContainer();
	if (pRoot == NULL) return FALSE;

	HWND hWndOwner = pRoot->GetOwner();
	ATLASSERT(IsWindow(hWndOwner));

	return ::ClientToScreen(hWndOwner, lpPoint);
}
BOOL CComponent::ClientToScreen(LPRECT lpRect)
{
	CRootContainer* pRoot = GetRootContainer();
	if (pRoot == NULL) return FALSE;

	HWND hWndOwner = pRoot->GetOwner();
	ATLASSERT(IsWindow(hWndOwner));

	if(!::ClientToScreen(hWndOwner, (LPPOINT)lpRect))
		return FALSE;
	return ::ClientToScreen(hWndOwner, ((LPPOINT)lpRect)+1);
}
BOOL CComponent::ScreenToClient(LPPOINT lpPoint)
{
	CRootContainer* pRoot = GetRootContainer();
	if (pRoot == NULL) return FALSE;

	HWND hWndOwner = pRoot->GetOwner();
	ATLASSERT(IsWindow(hWndOwner));

	return ::ScreenToClient(hWndOwner, lpPoint);
}
BOOL CComponent::ScreenToClient(LPRECT lpRect)
{
	CRootContainer* pRoot = GetRootContainer();
	if (pRoot == NULL) return FALSE;

	HWND hWndOwner = pRoot->GetOwner();
	ATLASSERT(IsWindow(hWndOwner));

	if(!::ScreenToClient(hWndOwner, (LPPOINT)lpRect))
		return FALSE;
	return ::ScreenToClient(hWndOwner, ((LPPOINT)lpRect)+1);
}
BOOL CComponent::InvalidateImmediately()
{
	if (_IsWindowAccessible() == FALSE) return TRUE;

	CRootContainer* pRoot = GetRootContainer();
	if (pRoot == NULL) return FALSE;

	CWindow window = pRoot->GetOwner();
	ATLASSERT(window.IsWindow());

	CRect rcRepaint;
	GetBounds(&rcRepaint);
	POINT ptAbsolute;
	GetParentAbsolutePosition(&ptAbsolute);
	rcRepaint.OffsetRect(ptAbsolute);

	return window.RedrawWindow(rcRepaint);
}
void CComponent::SetVisible(BOOL bVisible)
{
	m_bVisible = bVisible;
	if (m_bVisible)
	{
		// {kr} 콺 Ʈ   {/kr}
		BOOL bIsMouseOver = this->IsMouseOver();
		if (bIsMouseOver == TRUE)
		{
			// {kr} 콺 Enter ȣ {/kr}
			POINT ptCur = {0, 0};
			this->OnMouseEnter(ptCur);
		}
		
		OnShowNotify();
	}
	else
	{
		// {kr} 콺 Leave ȣ {/kr}
		this->OnMouseLeave();
		OnHideNotify();
	}
}
BOOL CComponent::IsVisible()
{
	return m_bVisible;
}
BOOL CComponent::IsMouseOver()
{ 
	POINT ptCursor = {0, 0};
	GetCursorPos(&ptCursor);
	HWND hWndOwner = this->GetOwnerWindow();
	if (hWndOwner == NULL)
	{
		return FALSE;
	}
	ATLASSERT(::IsWindow(hWndOwner));

	::ScreenToClient(hWndOwner, &ptCursor);

	//hhdadd 2008/02/14
	//return PtInRect(ptCursor);

	CRect rcBound;
	GetBounds(&rcBound);
	//rcBound.right -= m_Margin.right;
	//rcBound.bottom -= m_Margin.bottom;

	//get absolute position
	POINT ptAbsolute;
	if (GetParentAbsolutePosition(&ptAbsolute))
	{
		rcBound.OffsetRect(ptAbsolute);		
	}
	return rcBound.PtInRect(ptCursor);
}
void CComponent::SetParent(CContainer* pParent)
{
	m_pParent = pParent;
}
CContainer*	CComponent::GetParent()
{
	return m_pParent;
}
BOOL CComponent::IsCaptured()
{
	CRootContainer* pRootContainer = GetRootContainer();
	ATLASSERT(pRootContainer != NULL);
	if (pRootContainer)
	{
		if (this == pRootContainer->GetCapturedComponent())
		{
			return TRUE;
		}
	}	
	return FALSE;
}
BOOL CComponent::PtInRect(POINT pt)
{		
	CRect rcBound;
	GetBounds(&rcBound);
	//rcBound.right -= m_Margin.right;
	//rcBound.bottom -= m_Margin.bottom;

	//get absolute position
	POINT ptAbsolute;
	if (GetParentAbsolutePosition(&ptAbsolute))
	{
		rcBound.OffsetRect(ptAbsolute);		
	}

	BOOL bSuccess = rcBound.PtInRect(pt);
	return bSuccess;
}
BOOL CComponent::HitTest(POINT pt)
{
	return PtInRect(pt);
}
BOOL CComponent::ParentPtInRect(POINT pt)
{
	if (this->m_pParent != NULL)
	{
		return this->m_pParent->PtInRect(pt);
	}
	return TRUE;
}
CComponent* CComponent::SetFocus()
{
	CRootContainer* pRootContainer = GetRootContainer();
	ATLASSERT(pRootContainer != NULL);
	if (pRootContainer)
	{
		if (_IsWindowAccessible() == FALSE) 
		{
			return pRootContainer->GetFocus();
		}
		return pRootContainer->SetFocus(this);
	}
	return NULL;
}
CComponent* CComponent::GetFocus()
{
	CRootContainer* pRootContainer = GetRootContainer();
	ATLASSERT(pRootContainer != NULL);
	if (pRootContainer)
	{
		return pRootContainer->GetFocus();
	}
	return NULL;
}
void CComponent::SetCapture()
{
	CRootContainer* pRoot = GetRootContainer();
	if (pRoot == NULL) return;

	HWND hWndOwner = pRoot->GetOwner();
	ATLASSERT(IsWindow(hWndOwner));

	::SetCapture(hWndOwner);
	pRoot->SetCapturedComponent(this);
}
void CComponent::ReleaseCapture()
{
	CRootContainer* pRoot = GetRootContainer();
	if (pRoot == NULL) return;
	pRoot->SetCapturedComponent(NULL);
	::ReleaseCapture();
}
LRESULT	CComponent::SendMessage(UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	CRootContainer* pRoot = GetRootContainer();
	if (pRoot == NULL) return 0;

	HWND hWndOwner = pRoot->GetOwner();
	ATLASSERT(IsWindow(hWndOwner));

	return ::SendMessage(hWndOwner, uMsg, wParam, lParam);
}
LRESULT	CComponent::PostMessage(UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	CRootContainer* pRoot = GetRootContainer();
	if (pRoot == NULL) return 0;

	HWND hWndOwner = pRoot->GetOwner();
	ATLASSERT(IsWindow(hWndOwner));

	return ::PostMessage(hWndOwner, uMsg, wParam, lParam);
}
void CComponent::SetName(LPCTSTR pszcName)
{
	if (pszcName == NULL) return;
	ZeroMemory(m_szName,MAX_COMPONENT_NAME);
	_tcsncpy(m_szName, pszcName, _tcslen(pszcName));
}
CComponent* CComponent::FindName(LPCTSTR pszcName)
{
	if (_tcscmp(pszcName, m_szName) == 0) return this;
	return NULL;
}
CComponent* CComponent::FindParentName(LPCTSTR pszcName)
{
	return this->FindName(pszcName);
}
void CComponent::GetMargin(LPRECT margin)
{
	if (margin)
	{
		CopyRect(margin, &m_Margin);
	}
}
void CComponent::SetMargin(LPRECT margin)
{
	if (margin)
	{
		CopyRect(&m_Margin, margin);
	}
}
void CComponent::SetUI(IComponentUI* pComponentUI)
{
	m_pComponentUI = pComponentUI;
}
IComponentUI* CComponent::GetUI()
{
	return m_pComponentUI;
}
HWND CComponent::GetOwnerWindow()
{
	CRootContainer* pRoot = GetRootContainer();
	if (pRoot == NULL) return NULL;

	HWND hWndOwner = pRoot->GetOwner();
	return hWndOwner;
}
void CComponent::AddMouseListener(IMouseListener* pMouseListener)
{
	ATLASSERT(pMouseListener != NULL);
	ATLASSERT(m_vMouseListener.Find(pMouseListener) == NULL);

	m_vMouseListener.AddTail(pMouseListener);
}
void CComponent::RemoveMouseListener(IMouseListener* pMouseListener)
{
	ATLASSERT(pMouseListener != NULL);
	POSITION pos = m_vMouseListener.Find(pMouseListener);
	if (pos)
	{
		m_vMouseListener.RemoveAt(pos);		
	}
}
void CComponent::AddKeyListener(IKeyListener* pKeyListener)
{
	ATLASSERT(pKeyListener != NULL);
	ATLASSERT(m_vKeyListener.Find(pKeyListener) == NULL);

	m_vKeyListener.AddTail(pKeyListener);
}
void CComponent::RemoveKeyListener(IKeyListener* pKeyListener)
{
	ATLASSERT(pKeyListener != NULL);
	POSITION pos = m_vKeyListener.Find(pKeyListener);
	if (pos)
	{
		m_vKeyListener.RemoveAt(pos);		
	}
}
void CComponent::AddPropertyChangedListener(IPropertyChangedListener* pPropertyChangedListener)
{
	ATLASSERT(pPropertyChangedListener != NULL);
	ATLASSERT(m_vPropertyListener.Find(pPropertyChangedListener) == NULL);

	m_vPropertyListener.AddTail(pPropertyChangedListener);
	pPropertyChangedListener->AddRef();
}
void CComponent::RemovePropertyChangedListener(IPropertyChangedListener* pPropertyChangedListener)
{
	ATLASSERT(pPropertyChangedListener != NULL);
	POSITION pos = m_vPropertyListener.Find(pPropertyChangedListener);
	if (pos)
	{
		m_vPropertyListener.RemoveAt(pos);
		pPropertyChangedListener->AutoRelease();
	}
}
void CComponent::FirePropertyChangedEvent(CPropertyChangeEvent* pPropertyEvent)
{
	POSITION pos = m_vPropertyListener.GetHeadPosition();
	while (pos)
	{
		IPropertyChangedListener* pPropertyChangedListener = m_vPropertyListener.GetNext(pos);
		if (pPropertyChangedListener)
		{
			pPropertyChangedListener->OnPropertyChanged(pPropertyEvent);
		}
	}
}
Binding* CComponent::CreateBinding()
{
	return new Binding;
}
void CComponent::SetBinding(LPCTSTR propertyName, Binding* binding)
{
	ATLASSERT(binding != NULL);

	binding->SetTarget(this, propertyName);
	binding->Sync();
	m_vBinding.AddTail(binding);
}
void CComponent::_DestroyBinding()
{
	POSITION pos = m_vBinding.GetHeadPosition();
	while (pos)
	{
		Binding* binding = m_vBinding.GetNext(pos);
		if (binding)
		{
			binding->AutoRelease();
		}
	}	
}
void CComponent::OnLButtonDblClk(UINT nFlags, CPoint point) 
{
	if(this->IsWindowEnabled() == FALSE) return;

	CMouseEvent mouseevent(this, CMouseEvent::E_MOUSE_BUTTON_LEFT, 2, nFlags, point);
	POSITION pos = m_vMouseListener.GetHeadPosition();
	while (pos)
	{
		IMouseListener* pMouseListener = m_vMouseListener.GetNext(pos);
		if (pMouseListener)
		{
			pMouseListener->OnMousePressed(&mouseevent);
		}
	}
}
void CComponent::OnLButtonDown(UINT nFlags, CPoint point) 
{
	if(this->IsWindowEnabled() == FALSE) return;

	CMouseEvent mouseevent(this, CMouseEvent::E_MOUSE_BUTTON_LEFT, 1, nFlags, point);
	POSITION pos = m_vMouseListener.GetHeadPosition();
	while (pos)
	{
		IMouseListener* pMouseListener = m_vMouseListener.GetNext(pos);
		if (pMouseListener)
		{
			pMouseListener->OnMousePressed(&mouseevent);
		}
	}
}
void CComponent::OnLButtonUp(UINT nFlags, CPoint point) 
{
	CMouseEvent mouseevent(this, CMouseEvent::E_MOUSE_BUTTON_LEFT, 1, nFlags, point);
	POSITION pos = m_vMouseListener.GetHeadPosition();
	while (pos)
	{
		IMouseListener* pMouseListener = m_vMouseListener.GetNext(pos);
		if (pMouseListener)
		{
			pMouseListener->OnMouseReleased(&mouseevent);
		}
	}
}
void CComponent::OnRButtonDblClk(UINT nFlags, CPoint point) 
{
	if(this->IsWindowEnabled() == FALSE) return;

	CMouseEvent mouseevent(this, CMouseEvent::E_MOUSE_BUTTON_RIGHT, 2, nFlags, point);
	POSITION pos = m_vMouseListener.GetHeadPosition();
	while (pos)
	{
		IMouseListener* pMouseListener = m_vMouseListener.GetNext(pos);
		if (pMouseListener)
		{
			pMouseListener->OnMousePressed(&mouseevent);
		}
	}
}
void CComponent::OnRButtonDown(UINT nFlags, CPoint point) 
{
	if(this->IsWindowEnabled() == FALSE) return;

	CMouseEvent mouseevent(this, CMouseEvent::E_MOUSE_BUTTON_RIGHT, 1, nFlags, point);
	POSITION pos = m_vMouseListener.GetHeadPosition();
	while (pos)
	{
		IMouseListener* pMouseListener = m_vMouseListener.GetNext(pos);
		if (pMouseListener)
		{
			pMouseListener->OnMousePressed(&mouseevent);
		}
	}
}
void CComponent::OnRButtonUp(UINT nFlags, CPoint point) 
{
	if(this->IsWindowEnabled() == FALSE) return;

	CMouseEvent mouseevent(this, CMouseEvent::E_MOUSE_BUTTON_RIGHT, 1, nFlags, point);
	POSITION pos = m_vMouseListener.GetHeadPosition();
	while (pos)
	{
		IMouseListener* pMouseListener = m_vMouseListener.GetNext(pos);
		if (pMouseListener)
		{
			pMouseListener->OnMouseReleased(&mouseevent);
		}
	}
}
void CComponent::OnMouseMove(UINT nFlags, CPoint point)
{
	if(this->IsWindowEnabled() == FALSE) return;

	CRootContainer* pRoot = GetRootContainer();
	if (pRoot == NULL) return;
	pRoot->SetLastCursorComponent(this);

	CMouseEvent mouseevent(this, CMouseEvent::E_MOUSE_BUTTON_NONE, -1, nFlags, point);
	POSITION pos = m_vMouseListener.GetHeadPosition();
	while (pos)
	{
		IMouseListener* pMouseListener = m_vMouseListener.GetNext(pos);
		if (pMouseListener)
		{
			pMouseListener->OnMouseMoved(&mouseevent);
		}
	}
}
void CComponent::OnMouseLeave()
{
	if (m_bEnter == FALSE) return;

	if(this->_IsWindowAccessible() == FALSE) return;

	CMouseEvent mouseevent(this, CMouseEvent::E_MOUSE_BUTTON_NONE, -1, (UINT)(-1), CPoint());
	POSITION pos = m_vMouseListener.GetHeadPosition();
	while (pos)
	{
		IMouseListener* pMouseListener = m_vMouseListener.GetNext(pos);
		if (pMouseListener)
		{
			pMouseListener->OnMouseExited(&mouseevent);
		}
	}
	m_bEnter = FALSE;
}
void CComponent::OnMouseEnter(CPoint point)
{
	if (m_bEnter == TRUE) return ;
	if(this->_IsWindowAccessible() == FALSE) return;

	CMouseEvent mouseevent(this, CMouseEvent::E_MOUSE_BUTTON_NONE, -1, (UINT)(-1), point);
	POSITION pos = m_vMouseListener.GetHeadPosition();
	while (pos)
	{
		IMouseListener* pMouseListener = m_vMouseListener.GetNext(pos);
		if (pMouseListener)
		{
			pMouseListener->OnMouseEntered(&mouseevent);
		}
	}
	m_bEnter = TRUE;
}
void CComponent::OnMouseWheel(UINT nFlags, short nDelta, CPoint point)
{
	if(this->IsWindowEnabled() == FALSE) return;

	CMouseEvent mouseevent(this, CMouseEvent::E_MOUSE_BUTTON_NONE, -1, nFlags, point, nDelta);
	POSITION pos = m_vMouseListener.GetHeadPosition();
	while (pos)
	{
		IMouseListener* pMouseListener = m_vMouseListener.GetNext(pos);
		if (pMouseListener)
		{
			pMouseListener->OnMouseWheelMoved(&mouseevent);
		}
	}
}
void CComponent::OnKeyDown(UINT nChar, UINT nRepCount, UINT nFlags)
{
	if(this->IsWindowEnabled() == FALSE) return;

	CKeyEvent keyevent(this, nChar, nRepCount, nFlags);
	POSITION pos = m_vKeyListener.GetHeadPosition();
	while (pos)
	{
		IKeyListener* pKeyListener = m_vKeyListener.GetNext(pos);
		if (pKeyListener)
		{
			pKeyListener->OnKeyPressed(&keyevent);
		}
	}
}
void CComponent::OnKeyUp(UINT nChar, UINT nRepCount, UINT nFlags)
{
	if(this->IsWindowEnabled() == FALSE) return;

	CKeyEvent keyevent(this, nChar, nRepCount, nFlags);
	POSITION pos = m_vKeyListener.GetHeadPosition();
	while (pos)
	{
		IKeyListener* pKeyListener = m_vKeyListener.GetNext(pos);
		if (pKeyListener)
		{
			pKeyListener->OnKeyReleased(&keyevent);
		}
	}
}
LRESULT CComponent::OnMessage(UINT uMsg, WPARAM wParam, LPARAM lParam, BOOL& bHandled)
{
	LRESULT lResult = FALSE;

	bHandled = FALSE;
	return lResult;
}
#define IMPLEMENT_MOUSE_HANDLER(_fn_) \
	CPoint ptMouse(lParam); \
	if (IsCaptured() || (ParentPtInRect(ptMouse) && HitTest(ptMouse))) \
	{ \
		POINT ptAbsolute; \
		GetParentAbsolutePosition(&ptAbsolute); \
		CRect rcBound;\
		GetBounds(&rcBound);\
		ptMouse.x -= rcBound.left + ptAbsolute.x; \
		ptMouse.y -= rcBound.top + ptAbsolute.y; \
		this->_fn_(static_cast<UINT>(wParam), ptMouse); \
	} else \
	{ \
		bHandled = FALSE; \
	}
IMPLEMENT_MESSAGE_HANDLER(CComponent, OnLButtonDblClk)
{
	IMPLEMENT_MOUSE_HANDLER(OnLButtonDblClk);
	return 0;
}
IMPLEMENT_MESSAGE_HANDLER(CComponent, OnLButtonDown)
{
	IMPLEMENT_MOUSE_HANDLER(OnLButtonDown);
	return 0;
}
IMPLEMENT_MESSAGE_HANDLER(CComponent, OnLButtonUp)
{
	IMPLEMENT_MOUSE_HANDLER(OnLButtonUp);
	return 0;
}
IMPLEMENT_MESSAGE_HANDLER(CComponent, OnRButtonDblClk)
{
	IMPLEMENT_MOUSE_HANDLER(OnRButtonDblClk);
	return 0;
}
IMPLEMENT_MESSAGE_HANDLER(CComponent, OnRButtonDown)
{
	IMPLEMENT_MOUSE_HANDLER(OnRButtonDown);
	return 0;
}
IMPLEMENT_MESSAGE_HANDLER(CComponent, OnRButtonUp)
{
	IMPLEMENT_MOUSE_HANDLER(OnRButtonUp);
	return 0;
}
IMPLEMENT_MESSAGE_HANDLER(CComponent, OnMouseMove)
{
 //	CRootContainer* pRoot = GetRootContainer();
 //	if (pRoot != NULL)
 //	{
 //		HWND hWndOwner = pRoot->GetOwner();
 //		if(hWndOwner)
 //		{
 //			TRACKMOUSEEVENT TrackEvent;
 //			TrackEvent.cbSize      = sizeof( TRACKMOUSEEVENT );
 //			TrackEvent.dwFlags     = TME_LEAVE;
 //			TrackEvent.hwndTrack   = hWndOwner;
 //			TrackEvent.dwHoverTime = 1;
 //			TrackMouseEvent( &TrackEvent );
 //		}
 //	}

	CPoint ptMouse(lParam);
	if (IsCaptured() || (ParentPtInRect(ptMouse) && HitTest(ptMouse)))
	{
		POINT ptAbsolute;
		GetParentAbsolutePosition(&ptAbsolute);
		CRect rcBound;
		GetBounds(&rcBound);
		ptMouse.x -= rcBound.left + ptAbsolute.x;
		ptMouse.y -= rcBound.top + ptAbsolute.y;
		this->OnMouseMove(static_cast<UINT>(wParam), ptMouse);
		// {kr} 콺Ͱ Ʈѷ ó ´ٸ. {/kr}
		if (m_bEnter == FALSE)
		{
			// {kr} mouse enter ȣ . {/kr}
			this->OnMouseEnter(ptMouse);			
		}
	}
	else
	{
		if (m_bEnter == TRUE)
		{
			this->OnMouseLeave();			
		}
		bHandled = FALSE;
	}
	return 0;
}
IMPLEMENT_MESSAGE_HANDLER(CComponent, OnMouseWheel)
{
	CPoint ptMouse(lParam);
	CRootContainer* pRoot = GetRootContainer();
	if (pRoot == NULL) 
	{
		return 0;
	}
	HWND hWndOwner = pRoot->GetOwner();
	CRect rc;
	pRoot->GetBounds(&rc);
	::ScreenToClient(hWndOwner,&ptMouse);
	if(!rc.PtInRect(ptMouse))
	{
		return 0;
	}
	if (PtInRect(ptMouse))
	{
		POINT ptAbsolute;
		GetParentAbsolutePosition(&ptAbsolute);
		CRect rcBound;
		GetBounds(&rcBound);
		ptMouse.x -= rcBound.left + ptAbsolute.x;
		ptMouse.y -= rcBound.top + ptAbsolute.y;
		this->OnMouseWheel(static_cast<UINT>(GET_KEYSTATE_WPARAM(wParam)), 
			static_cast<short>(GET_WHEEL_DELTA_WPARAM(wParam)), 
			ptMouse);
	}
	else
	{
		bHandled = FALSE;
	}
	return 0;
}
IMPLEMENT_MESSAGE_HANDLER(CComponent, OnKeyDown)
{
	CComponent* pFocusedComponent = GetFocus();
	if (pFocusedComponent == this)
	{
		this->OnKeyDown(static_cast<UINT>(wParam), (UINT)lParam & 0xFFFF, (UINT)((lParam & 0xFFFF0000) >> 16));
	}
	else
	{
		bHandled = FALSE;
	}
	return 0;
}
IMPLEMENT_MESSAGE_HANDLER(CComponent, OnKeyUp)
{
	CComponent* pFocusedComponent = GetFocus();
	if (pFocusedComponent == this)
	{
		this->OnKeyUp(static_cast<UINT>(wParam), (UINT)lParam & 0xFFFF, (UINT)((lParam & 0xFFFF0000) >> 16));
	}
	else
	{
		bHandled = FALSE;
	}
	return 0;
}
IMPLEMENT_MESSAGE_HANDLER(CComponent, OnSetCursor)
{	
	HWND hWndOwner = GetOwnerWindow();
	CPoint ptMouse;
	GetCursorPos(&ptMouse);
	::ScreenToClient(hWndOwner, &ptMouse);

	if (PtInRect(ptMouse))
	{
		OnSetCursor(HIWORD(lParam));
	}
	else
	{
		bHandled = FALSE;
	}
	return 0;
}
IMPLEMENT_MESSAGE_HANDLER(CComponent, OnUserDefinedMessage)
{
	this->OnMessage(uMsg, wParam, lParam, bHandled);
	return 0;
}
//IMPLEMENT_MESSAGE_HANDLER(CComponent, OnMouseLeave)
//{
//	this->OnMouseAway();
//
//	bHandled = FALSE;
//	return 0;
//}
BOOL CComponent::GetParentAbsolutePosition(LPPOINT pPt)
{	
	pPt->x = 0;
	pPt->y = 0;
	//get absolute position
	POINT ptScroll = {0, 0};
	CRect rcBound;
	CRect rcMargin;
	rcBound.SetRectEmpty();
	CContainer* pParent = m_pParent;
	CComponent* pPreParent = this;
	while (pParent)
	{
		pParent->GetBounds(rcBound);
		
		pPt->x += rcBound.left;
		pPt->y += rcBound.top;

		if (pPreParent->IsNCComponent() == FALSE)
		{
			if (pParent->_GetScrollPos(&ptScroll))
			{
				pPt->x -= ptScroll.x;
				pPt->y -= ptScroll.y;
			}
		}
		pPreParent = pParent;
		pParent = pParent->GetParent();
	}
	return TRUE;
}
CRootContainer*	CComponent::GetRootContainer()
{
	CRootContainer* pRootContainer = NULL;
	if (IS_EXACT_CLASS(CRootContainer, this))
	{
		pRootContainer = reinterpret_cast<CRootContainer*>(this);
		return pRootContainer;
	}

	CContainer* pParent = m_pParent;
	while (pParent)
	{
		if (IS_EXACT_CLASS(CRootContainer, pParent))
		{
			pRootContainer = reinterpret_cast<CRootContainer*>(pParent);
			break;
		}
		pParent = reinterpret_cast<CContainer*>(pParent->GetParent());
	}
	return pRootContainer;
}
BOOL CComponent::GetPreferredSize(LPSIZE lpSize)
{
	BOOL bSuccess = GetOrginPreferredSize(lpSize);
	/*if (bSuccess)
	{
		lpSize->cx += (m_Margin.right);
		lpSize->cy += (m_Margin.bottom);
	}*/
	return bSuccess;
}
BOOL CComponent::GetOrginPreferredSize(LPSIZE lpSize)
{
	CRect rcBound;	
	rcBound.left = GetRealValue(m_rcBound.left);
	rcBound.top = GetRealValue(m_rcBound.top);
	rcBound.right = GetRealValue(m_rcBound.right, rcBound.left);
	rcBound.bottom = GetRealValue(m_rcBound.bottom, rcBound.top);

	lpSize->cx = rcBound.Width();
	lpSize->cy = rcBound.Height();

	return ( rcBound.Width() > 0 || rcBound.Height() > 0 );
}
void CComponent::_RemoveParent()
{
	if (m_pParent)
	{
		m_pParent->Remove(this);
		m_pParent = NULL;
	}
}
BOOL CComponent::IsNCComponent() 
{ 
	CContainer* pParent = m_pParent;
	if (pParent)
	{
		if (pParent->ContainedNCComponent(this))
		{
			return TRUE;
		}
		if (pParent->IsNCComponent())
		{
			return TRUE;
		}		
	}
	return FALSE;
}

void CComponent::_Monitor_CursorEnterExit( LPRECT prcBoundOld, LPRECT prcBoundNew )
{
	if (this->GetRootContainer() == NULL) return;
	if (this->GetOwnerWindow() == NULL) return;
	
	// {kr} Ʈ ̰, 콺 ʹ ɶ  {/kr}
	// {kr} Ʈ /  {/kr}
	CRect rcBoundOld(prcBoundOld);
	CRect rcBoundNew(prcBoundNew);

	// {kr}  콺 ͸  {/kr}
	POINT ptCur;
	::GetCursorPos(&ptCur);
	this->ScreenToClient(&ptCur);

	// {kr} /    콺 Ͱ over ƴ Ǵ. {/kr}
	POINT ptAbsolute;
	if (GetParentAbsolutePosition(&ptAbsolute))
	{
		rcBoundOld.OffsetRect(ptAbsolute);
		rcBoundNew.OffsetRect(ptAbsolute);
	}
	BOOL bPtInRectOld = rcBoundOld.PtInRect(ptCur);
	BOOL bPtInRectNew = rcBoundNew.PtInRect(ptCur);

	// {kr} ſ over ¿ٰ 翡 over ° ƴϸ {/kr}
	if (bPtInRectOld == TRUE &&
		bPtInRectNew == FALSE)
	{
		// {kr} exit ̺Ʈ ߻ {/kr}
		CPropertyChangeEvent PropEvent(this, ComponentProperty::MouseExit);
		this->FirePropertyChangedEvent(&PropEvent);
	}

	// {kr} ſ over ° ƴϴٰ 翡 over ̸ {/kr}
	if (bPtInRectOld == FALSE &&
		bPtInRectNew == TRUE)
	{
		// {kr} enter ̺Ʈ ߻ {/kr}
		CPropertyChangeEvent PropEvent(this, ComponentProperty::MouseExit);
		this->FirePropertyChangedEvent(&PropEvent);
	}
}
